-- operate vmebus with a file
-- file format is am(hex) adr(hex) data(hex) r|w b|w|q s|b [opt burst count,decimal]
-- b = byte, w=word, q=quad
-- s = single cycle, b = burst
-- in burst mode, follow with count of data words, up to 256
library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;
use std.textio.all;

entity vmebusop is
    port (
		vme_adr : out std_logic_vector(31 downto 0);
		vme_data : out std_logic_vector(31 downto 0);
		vme_not_rw : out std_logic;
		vme_not_as,vme_not_ds0,vme_not_ds1 : in std_logic; -- use to tell when cycle is done
	vme_am : out std_logic_vector(5 downto 0);
	vme_mode : out std_logic_vector(1 downto 0); -- 1 - byte, 2-two byte, 3- quad byte
	vme_burst : out std_logic; -- 1 run burst mode
	vme_burst_cnt : out std_logic_vector(7 downto 0); -- burst count
	vme_go : out std_logic :='0'
		   );
end vmebusop;

library ieee;
use ieee.std_logic_1164.all;
library exemplar;
use exemplar.exemplar_1164.all;
use std.textio.all;

architecture operate_vmebusop of vmebusop is

type integer_array is array(255 downto 0) of integer;

procedure vmebop(
  file romdef : TEXT;
  amv,adrv  : inout integer;
  datav : inout integer_array;
  variable rwmode : inout std_logic;
  variable opmode : inout std_logic_vector(1 downto 0);
  variable burst : inout std_logic;
  variable burst_cnt : inout std_logic_vector(7 downto 0);
  variable flag : inout boolean
  ) is
variable l : line;
variable j : integer;
variable chr : character;
variable sptr : integer :=0;
variable forever : integer :=16000000;
variable bcnt : integer;


        type hex_to_v4_t is array(character) of integer;
        constant hex_to_v4 : hex_to_v4_t := (
            '0' => 0,
            '1' => 1,
            '2' => 2,
            '3' => 3,
            '4' => 4,
            '5' => 5,
            '6' => 6,
            '7' => 7,
            '8' => 8,
            '9' => 9,
            'a' | 'A' => 10,
            'b' | 'B' => 11,
            'c' | 'C' => 12,
            'd' | 'D' => 13,
            'e' | 'E' => 14,
            'f' | 'F' => 15,
            others => -1);

        -------------------------------------------
        -- convert a hex string to integer
        -------------------------------------------
        procedure hex_str_to_v( s : in string; ptr : inout integer; result : inout integer:=0; flag : inout boolean)  is
            variable mult : integer :=1;
        begin
			flag := false;
			result :=0;
            for i in ptr to s'high loop
			    ptr := i;
			    if ((hex_to_v4(s(i)) = -1) and (flag)) then exit; end if; -- hit terminator
                result := result*16;
			    if (hex_to_v4(s(i)) /= -1) then
				  flag := true;
                  result := result + hex_to_v4(s(i));
				end if;
            end loop;
        end procedure;

        type dec_to_v4_t is array(character) of integer;
        constant dec_to_v4 : dec_to_v4_t := (
            '0' => 0,
            '1' => 1,
            '2' => 2,
            '3' => 3,
            '4' => 4,
            '5' => 5,
            '6' => 6,
            '7' => 7,
            '8' => 8,
            '9' => 9,
            others => -1);

        -------------------------------------------
        -- convert a decimal string to integer
        -------------------------------------------
        procedure dec_str_to_v( s : in string; ptr : inout integer; result : inout integer:=0; flag : inout boolean)  is
            variable mult : integer :=1;
        begin
			flag := false;
			result :=0;
            for i in ptr to s'high loop
			    ptr := i;
			    if ((dec_to_v4(s(i)) = -1) and (flag)) then exit; end if; -- hit terminator
                result := result*10;
			    if (dec_to_v4(s(i)) /= -1) then
				  flag := true;
                  result := result + dec_to_v4(s(i));
				end if;
            end loop;
        end procedure;

        procedure get_mode( s : in string; ptr : inout integer; variable mode : inout std_logic; flag : inout boolean)  is
        begin
			flag := false;
            for i in ptr to s'high loop
			    ptr := i;
				if (s(i) = 'r') then
				  mode := '1';
				  flag := true;
				  exit;
				elsif (s(i) = 'w') then
				  mode := '0';
				  flag := true;
				  exit;
				end if;
            end loop;
            ptr := ptr+1;
        end procedure;

        procedure get_op_mode( s : in string; ptr : inout integer; variable mode : inout std_logic_vector(1 downto 0); flag : inout boolean)  is
        begin
			flag := false;
            for i in ptr to s'high loop
			    ptr := i;
				if (s(i) = 'b') then
				  mode := "01";
				  flag := true;
				  exit;
				elsif (s(i) = 'w') then
				  mode := "10";
				  flag := true;
				  exit;
				elsif (s(i) = 'q') then
				  mode := "11";
				  flag := true;
				  exit;
				end if;
            end loop;
            ptr := ptr+1;
        end procedure;

        procedure get_burst_mode( s : in string; ptr : inout integer; variable mode : inout std_logic; flag : inout boolean)  is
        begin
			flag := false;
            for i in ptr to s'high loop
			    ptr := i;
				if (s(i) = 's') then
				  mode := '0';
				  flag := true;
				  exit;
				elsif (s(i) = 'b') then
				  mode := '1';
				  flag := true;
				  exit;
				end if;
            end loop;
            ptr := ptr+1;
        end procedure;
begin
    readline(romdef,l);
	sptr :=1;
    hex_str_to_v(l.all,sptr,amv,flag);
    if not flag then return; end if;
    hex_str_to_v(l.all,sptr,adrv,flag);
    if not flag then return; end if;
    hex_str_to_v(l.all,sptr,datav(0),flag);
    if not flag then return; end if;
    get_mode(l.all,sptr,rwmode,flag);
    if not flag then return; end if;
    get_op_mode(l.all,sptr,opmode,flag);
    if not flag then return; end if;
    get_burst_mode(l.all,sptr,burst,flag);
    if not flag then return; end if;
	if (burst='1') then
	  dec_str_to_v(l.all,sptr,bcnt,flag);
      if not flag then return; end if;
	  burst_cnt := int2evec(bcnt)(7 downto 0);
	  FOR i IN 1 to bcnt-1 LOOP
        readline(romdef,l);
	    sptr :=1;
        hex_str_to_v(l.all,sptr,datav(i),flag);
        if not flag then exit; end if;
	  END LOOP;  
	end if;
end  vmebop;


constant vmebusstr  : string := "vmebus.tab";
FILE vmebusfile : TEXT open read_mode is vmebusstr;  
constant end_cycle_delay : time := 100 ns;

begin

proc_vmebop: process 
variable  amv,adrv : integer;
variable datav : integer_array;
variable  rwmode : std_logic;
variable  opmode : std_logic_vector(1 downto 0);
variable  burst : std_logic;
variable  burst_cnt : std_logic_vector(7 downto 0);
variable bcnt : integer;
variable flag : boolean;
begin
    vme_go <= '0';
    wait for 500 ns;
	while (true) loop
	vmebop(vmebusfile,amv,adrv,datav,rwmode,opmode,burst,burst_cnt,flag);
    if not flag then exit; end if;
	vme_am <= int2evec(amv)(5 downto 0);
	vme_adr <= int2evec(adrv)(31 downto 0);
	vme_data <= int2evec(datav(0))(31 downto 0);
	vme_mode <= opmode;
	vme_not_rw <= rwmode;
	vme_burst <= burst;
	vme_burst_cnt <= burst_cnt+"11111111";
    vme_go <= '1';
    if (burst='1') then
      bcnt := evec2int(burst_cnt);
      for i in 0 to bcnt-1 loop
        wait until rising_edge(vme_not_ds0) or rising_edge(vme_not_ds1);
        if not(i=(bcnt-1)) then
	      vme_data <= int2evec(datav(i+1))(31 downto 0);
        end if;
      end loop;
    end if;
    wait until rising_edge(vme_not_as);
    vme_go <= '0';
    if endfile(vmebusfile) then exit; end if;
    wait for end_cycle_delay;
    end loop;
	file_close(vmebusfile);
    while (true) loop wait for 1 ns; end loop;
end process proc_vmebop;

end operate_vmebusop;

